from .util import normalize
from md2gemini import md2gemini


def f(md):
    return normalize(md2gemini(md, strip_html=True))


def test_keep_bold():
    md = "Sentence with **bold** in it."
    gem = "Sentence with **bold** in it."
    assert f(md) == gem


def test_keep_bold_2():
    md = "Sentence with __bold__ in it."
    gem = "Sentence with **bold** in it."
    assert f(md) == gem


def test_keep_italics():
    md = "Sentence with *italics* in it."
    gem = "Sentence with *italics* in it."
    assert f(md) == gem


def test_keep_italics_2():
    md = "Sentence with _italics_ in it."
    gem = "Sentence with *italics* in it."
    assert f(md) == gem


def test_remove_inline_html():
    md = "<i>test1</i> <b>test2</b> <em>test3</em> <made-up>test4</made-up>"
    gem = "test1 test2 test3 test4"
    assert f(md) == gem


def test_remove_html_block():
    md = "<div>\n<tag>html is here</tag>\n</div>\n\nText after"
    gem = "Text after"
    assert f(md) == gem



from .util import normalize
from md2gemini import md2gemini


def f(md):
    return normalize(md2gemini(md, link_func=(lambda x: "foo")))


def test_func_root_url():
    md = "[test](/url)"
    gem = "=> foo test"
    assert f(md) == gem


def test_func_absolute_url():
    md = "[test](https://duck.com/test)"
    gem = "=> foo test"
    assert f(md) == gem



from .util import normalize
from md2gemini import md2gemini


def f(md, links):
    return normalize(md2gemini(md, links=links))


def test_paragraph_links():
    md = """
* [foo](foo.md)

# headline

* [bar](bar.md)
    """

    gem = """
* foo[1]

=> foo.md 1: foo.md

# headline

* bar[2]

=> bar.md 2: bar.md
""".strip()

    assert f(md, links="paragraph") == gem


def test_links_in_lists():
    # https://github.com/makeworld-the-better-one/md2gemini/issues/30

    md = "* 1before [bar](/foo) after \n* next item"
    gem = """
* 1before 
=> /foo bar
after
* next item
    """.strip()

    assert f(md, "newline") == gem

    md = "* 2before [bar](/foo) after \n* next item"
    gem = """
* 2before bar after
* next item

=> /foo bar
    """.strip()

    assert f(md, "copy") == gem


def test_many_footnote_links():
    md = """
* [foo](foo.md)

# headline

* [bar](bar.md)
* [bar2](bar2.md)

Some text after

* [foo2](foo2.md)
    """

    gem = """
* foo[1]

=> foo.md 1: foo.md

# headline

* bar[2]
* bar2[3]

=> bar.md 2: bar.md
=> bar2.md 3: bar2.md

Some text after

* foo2[4]

=> foo2.md 4: foo2.md
""".strip()

    assert f(md, links="paragraph") == gem


def test_link_in_quote():
    # https://github.com/makeworld-the-better-one/md2gemini/issues/44

    md = """
Text before

> quote with [link](https://example.com) in it

Text after
    """

    gem = """
Text before

> quote with link[1] in it

=> https://example.com 1: https://example.com

Text after
    """.strip()

    assert f(md, links="paragraph") == gem


def test_link_in_quote_newline():
    # https://github.com/makeworld-the-better-one/md2gemini/issues/44

    md = """
Text before

> quote with newline link in it
>
> [link](https://example.com)

Text after
    """

    gem = """
Text before

> quote with newline link in it
> 
> link[1]

=> https://example.com 1: https://example.com

Text after
    """.strip()

    assert f(md, links="paragraph") == gem



from .util import normalize
from md2gemini import md2gemini


def f(md):
    return normalize(md2gemini(md, base_url="https://example.com/"))


def test_root_url():
    md = "[test](/url)"
    gem = "=> https://example.com/url test"
    assert f(md) == gem


def test_absolute_url():
    md = "[test](https://duck.com/test)"
    gem = "=> https://duck.com/test test"
    assert f(md) == gem



from .util import normalize
from md2gemini import md2gemini


def f(md):
    return normalize(md2gemini(md, md_links=True))


def test_markdown_path():
    md = "[test](foo.md)"
    gem = "=> foo.gmi test"
    assert f(md) == gem


def test_markdown_path_anchor():
    md = "[test](foo.md#bar)"
    gem = "=> foo.gmi test"
    assert f(md) == gem



import mistune
from .renderers import GeminiRenderer, NEWLINE, PARAGRAPH_DELIM, LINK_DELIM, LINEBREAK
import argparse
import sys
import os
import re


def __text_between(text, delim, n=0):
    """Get the text between two delimeters.

    n: Which occurence of the delimeter pair to act on, zero-indexed.
    """

    return text.split(delim)[n + 1]


def __replace_between(text, delim, new_text, n=0):
    """Replace the text between two delimeters, and get the full text back.

    n: Which occurence of the delimeter pair to act on, zero-indexed.
    """

    start = delim.join(text.split(delim)[: n + 1])
    end = delim.join(text.split(delim)[n + 2 :])
    # return start + delim + new_text + delim + end
    return start + new_text + end


def md2gemini(
    markdown,
    code_tag="",
    img_tag="[IMG]",
    indent=" ",
    ascii_table=False,
    frontmatter=False,
    jekyll=False,
    links="newline",
    plain=False,
    strip_html=False,
    base_url="",
    md_links=False,
    link_func=None,
    table_tag="table",
    checklist=True,
):
    """Convert the provided markdown text to the gemini format.
    code_tag: The default alt text for code blocks.

    img_tag: The text added after an image link, to indicate it's an image.

    indent: How much to indent sub-levels of a list. Put several spaces, or \\t for a tab.

    ascii_table: Use ASCII to create tables, not Unicode.

    frontmatter: Remove Jekyll and Zola style front matter before converting.

    jekyll: Skip jekyll frontmatter when processing - DEPRECATED.

    links: Set to 'off' to turn off links, 'paragraph' to have footnotes at the end of each
    paragraph, or 'at-end' to have footnotes at the end of the document. You can also set it
    to 'copy' to put links that copy the inline link text after each paragraph. Not using this
    flag, or having any other value will result in regular, newline links.

    plain: Set to True to remove special markings from output that text/gemini doesn't support,
    like the asterisks for bold and italics, as well as inline HTML.

    strip_html: Strip all inline and block HTML from Markdown.

    base_url: All links starting with a slash will have this URL prepended to them.

    md_links: Convert all links to local files ending in .md to end with .gmi instead.

    link_func: Custom function to apply to links. This function takes a string containing the link
    URL as parameter, and should return the new link.

    table_tag: "The default alt text for table blocks."

    checklist: whether to support GitHub-style checklist list items: [ ] and [x]
    """

    if len(markdown) == 0:
        return ""

    # Pre processing
    # Remove frontmatter
    frontmatterExists = False
    if frontmatter:
        lines = markdown.strip().splitlines()
        if lines[0] == "---" or lines[0] == "+++":
            frontmatterExists = True
    elif jekyll:
        lines = markdown.strip().splitlines()
        if lines[0] == "---":
            frontmatterExists = True

    if frontmatterExists:
        md_lines = []
        for i, line in enumerate(lines[1:]):  # Skip first front matter line
            if ((frontmatter or jekyll) and line == "---") or (
                frontmatter and line == "+++"
            ):
                # End of frontmatter, add all the lines below it
                md_lines.extend(lines[i + 2 :])
                break
        # Turn it back into text
        if md_lines != []:
            # If md_lines was empty, then effectively no removal would occur, and frontmatter would be processed
            markdown = NEWLINE.join(md_lines)

    # Conversion
    renderer = GeminiRenderer(
        code_tag=code_tag,
        img_tag=img_tag,
        indent=indent,
        ascii_table=ascii_table,
        links=links,
        plain=plain,
        strip_html=strip_html,
        base_url=base_url,
        md_links=md_links,
        link_func=link_func,
        table_tag=table_tag,
        checklist=checklist,
    )
    gem = mistune.create_markdown(
        escape=False, renderer=renderer, plugins=["table", "url", "task_lists"]
    )
    gemtext = gem(markdown)

    # Post processing

    # Remove newlines within paragraphs, and add two newlines after them.
    while PARAGRAPH_DELIM in gemtext:  # While there's still unprocessed paragraphs
        pg = __text_between(gemtext, PARAGRAPH_DELIM).strip()
        pg = pg.replace("\r\n", "\n")  # Make all newlines the same
        pg = pg.replace(
            "\n", " "
        )  # Get rid of newlines in the same paragraph, like markdown does
        pg += NEWLINE * 2  # Add a blank line between paragraphs
        gemtext = __replace_between(gemtext, PARAGRAPH_DELIM, pg)

    # Add in hard linebreaks
    gemtext = gemtext.replace(LINEBREAK, NEWLINE)

    # Add remaining footnotes at and of file
    gemtext += renderer._render_footnotes()

    # Remove double link delims, which are produced by multiple footnotes
    gemtext = gemtext.replace(LINK_DELIM + LINK_DELIM, LINK_DELIM)
    # Remove all the link delims that are next to newlines,
    # so that when link delims are replaced with newlines later in the code,
    # we don't end up with double newlines.
    gemtext = re.sub(
        r"((?<=\r\n)|(?<=\n)" + LINK_DELIM + r")|(" + LINK_DELIM + r"(?=\r\n|\n))",
        "",
        gemtext,
    )
    gemtext = gemtext.replace(LINK_DELIM, NEWLINE)

    # Remove left whitespace in the lines after links
    gemlines = gemtext.splitlines()
    if gemlines[-1] == "":
        gemlines = gemlines[:-1]
    pre = False  # Whether we're in a preformatted area or not
    length = len(gemlines)
    for i, line in enumerate(gemlines):
        # Maintain preformatted state
        if line.startswith("```"):
            pre = not pre
            continue
        if i + 1 < length - 1 and line.startswith("=>") and not pre:
            # It's a link, fix the next line by removing left whitespace
            gemlines[i + 1] = gemlines[i + 1].lstrip()
    gemtext = NEWLINE.join(gemlines)
    gemtext = gemtext.rstrip().lstrip("\r\n")

    return gemtext


# Main functions, for running as a script


def __convert_file(file, args):
    if file == sys.stdin:
        gem = md2gemini(
            file.read(),
            code_tag=args.code_tag,
            img_tag=args.img_tag,
            indent=args.indent,
            ascii_table=args.ascii_table,
            frontmatter=args.frontmatter,
            jekyll=args.jekyll,
            links=args.links,
            plain=args.plain,
            strip_html=args.strip_html,
            base_url=args.base_url,
            md_links=args.md_links,
            link_func=None,
            table_tag=args.table_tag,
            checklist=not args.no_checklist,
        )
        print(gem)
    else:
        with open(file, "r") as f:
            gem = md2gemini(
                f.read(),
                code_tag=args.code_tag,
                img_tag=args.img_tag,
                indent=args.indent,
                ascii_table=args.ascii_table,
                frontmatter=args.frontmatter,
                jekyll=args.jekyll,
                links=args.links,
                plain=args.plain,
                strip_html=args.strip_html,
                base_url=args.base_url,
                md_links=args.md_links,
                link_func=None,
                table_tag=args.table_tag,
                checklist=not args.no_checklist,
            )
        if args.write:
            newfile = os.path.splitext(os.path.basename(file))[0] + ".gmi"
            with open(os.path.join(args.dir, newfile), "w") as f:
                f.write(gem)
        else:
            print(gem)


def main():
    parser = argparse.ArgumentParser(
        description="Convert markdown to gemini.", prog="md2gemini"
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "file",
        nargs="*",
        help="Files to convert. If no files are specified then data will be read from stdin and printed to stdout.",
    )
    parser.add_argument(
        "-w",
        "--write",
        action="store_true",
        help="Write output to a new file of the same name, but with a .gmi extension.",
    )
    parser.add_argument(
        "-d", "--dir", help="The directory to write files to, if --write is used."
    )
    parser.add_argument(
        "-a",
        "--ascii-table",
        action="store_true",
        help="Use ASCII to create tables, not Unicode.",
    )
    parser.add_argument(
        "-f",
        "--frontmatter",
        action="store_true",
        help="Remove Jekyll and Zola style front matter before converting.",
    )
    parser.add_argument(
        "-j",
        "--jekyll",
        action="store_true",
        help="Skip jekyll frontmatter when processing - DEPRECATED.",
    )
    parser.add_argument(
        "--code-tag",
        type=str,
        help="What alt text to add to unlabeled code blocks. Defaults to empty string.",
    )
    parser.add_argument(
        "--img-tag",
        type=str,
        help="What text to add after image links. Defaults to '[IMG]'.\nWrite something like --img-tag='' to remove it.",
    )
    parser.add_argument(
        "--table-tag",
        type=str,
        help="What alt text to add to table blocks. Defaults to 'table'.\nWrite something like --table-tag='' to remove it.",
    )
    parser.add_argument(
        "-i",
        "--indent",
        type=str,
        help="The number of spaces to use for list indenting. Put 'tab' to use a tab instead.",
    )
    parser.add_argument(
        "-l",
        "--links",
        type=str,
        help="Set to 'off' to turn off links, 'paragraph' to have footnotes at the end of each paragraph, or 'at-end' to have footnotes at the end of the document. You can also set it to 'copy' to put links that copy the inline link text after each paragraph. Not using this flag, or having any other value will result in regular, newline links.",
    )
    parser.add_argument(
        "-p",
        "--plain",
        action="store_true",
        help="Remove special markings from output that text/gemini doesn't support, like the asterisks for bold and italics, and inline HTML",
    )
    parser.add_argument(
        "-s",
        "--strip-html",
        action="store_true",
        help="Strip all inline and block HTML from Markdown. Note that using --plain will strip inline HTML as well.",
    )
    parser.add_argument(
        "-b",
        "--base-url",
        type=str,
        help="All links starting with a slash will have this URL prepended to them.",
    )
    parser.add_argument(
        "-m",
        "--md-links",
        action="store_true",
        help="Convert all links to local files ending in .md to end with .gmi instead.",
    )
    parser.add_argument(
        "-c",
        "--no-checklist",
        action="store_true",
        help="Disable rendering of GitHub-style checklist list items: [ ] and [x]",
    )
    args = parser.parse_args()

    # Validation of command line args
    if args.write and args.dir is None:
        args.dir = "."
    if args.write and not os.path.isdir(args.dir):
        print("Directory", args.dir, "cannot be found.", file=sys.stderr)
        sys.exit(1)
    if args.indent == "tab":
        args.indent = "\t"
    elif not args.indent is None:
        try:
            args.indent = " " * int(args.indent)
        except ValueError:
            print(
                "Invalid indent value. Must be an integer, or 'tab'.", file=sys.stderr
            )
            sys.exit(1)
    if args.code_tag is None:
        args.code_tag = ""

    if args.img_tag is None:
        args.img_tag = "[IMG]"

    if args.table_tag is None:
        args.table_tag = "table"

    # If there aren't any files then read from stdin
    if args.file == []:
        __convert_file(sys.stdin, args)
        sys.exit(0)

    # Process each file sequentially
    for file in args.file:
        if not os.path.isfile(file):
            print("File", file, "cannot be found.", file=sys.stderr)
            sys.exit(1)
        __convert_file(file, args)


__all__ = ["GeminiRenderer", "md2gemini", "main", "NEWLINE", "__version__"]
__version__ = "1.9.1"



from .util import normalize
from md2gemini import md2gemini
import pytest

multiple_tags = pytest.fixture(params=["[IMG]", "IMG", "test", "AAA"])


@multiple_tags
def img_tag_no_text(request):
    md = "![](image.jpg)"
    gem = "=> image.jpg " + request.param
    assert normalize(md2gemini(md, img_tag=request.param)) == gem


def test_img_tag_no_text(img_tag_no_text):
    pass


@multiple_tags
def img_tag_with_text(request):
    md = "![text](image.jpg)"
    gem = "=> image.jpg text " + request.param
    assert normalize(md2gemini(md, img_tag=request.param)) == gem


def test_img_tag_with_text(img_tag_with_text):
    pass


def test_empty_img_tag_no_text():
    md = "![](image.jpg)"
    gem = "=> image.jpg"
    assert normalize(md2gemini(md, img_tag="")) == gem


def test_empty_img_tag_with_text():
    md = "![text](image.jpg)"
    gem = "=> image.jpg text"
    assert normalize(md2gemini(md, img_tag="")) == gem


@multiple_tags
def table_tag(request):
    md = """
a|b|c
-|-|
1|2|3
4|5|6
""".lstrip()
    gem = (
        "```"
        + request.param
        + """
┌───┬───┬───┐
│ a │ b │ c │
╞═══╪═══╪═══╡
│ 1 │ 2 │ 3 │
├───┼───┼───┤
│ 4 │ 5 │ 6 │
└───┴───┴───┘
```"""
    )
    assert normalize(md2gemini(md, table_tag=request.param)) == gem


def test_table_tag(table_tag):
    pass



from .util import normalize
from md2gemini import md2gemini


def f(md):
    return normalize(md2gemini(md, frontmatter=True))


def test_remove_frontmatter_jekyll():
    md = "---\nsome text\n---\nbeginning"
    gem = "beginning"
    assert f(md) == gem


def test_remove_frontmatter_zola():
    md = "+++\nsome text\n+++\nbeginning"
    gem = "beginning"
    assert f(md) == gem


def test_no_remove_frontmatter_jekyll():
    # "Frontmatter" doesn't have an end
    md = "---\nsome text"
    assert f(md).endswith("some text")


def test_no_remove_frontmatter_zola():
    # "Frontmatter" doesn't have an end
    md = "+++\nsome text"
    gem = "+++ some text"
    assert f(md) == gem



from .util import normalize
from md2gemini import md2gemini
import pytest


def f(md):
    return normalize(md2gemini(md))


def test_normal_list():
    md = "- hi\n- oh"
    gem = "* hi\n* oh"
    assert f(md) == gem


@pytest.fixture(params=[2, 3, 4, 5, 6])
def list_with_newline_and_spaces(request):
    md = "- hello\n" + (" " * request.param) + "world\n- oh"
    gem = "* hello world\n* oh"
    assert f(md) == gem


def test_list_with_newline_and_spaces(list_with_newline_and_spaces):
    pass


def test_separate_lists():
    md = "* hello\n\n- oh"
    gem = "* hello\n\n* oh"
    assert f(md) == gem



"""
All the renderers that convert markdown to gemini.
"""

import re
import mistune
from .unitable import UniTable, ArraySizeError


NEWLINE = "\r\n"  # For Windows support
PARAGRAPH_DELIM = "\x02"  # The marker for paragraph start and end, for post processing
LINK_DELIM = "\x03"
LINEBREAK = "\x01"  # Represents a hard linebreak that should not be changed

FENCE_EXPR = re.compile(r"^( *)```")


class GeminiRenderer(
    mistune.HTMLRenderer
):  # Actually BaseRenderer should be used but this isn't available

    # NAME = "gemini"

    def __init__(
        self,
        code_tag="",
        img_tag="[IMG]",
        indent=" ",
        ascii_table=False,
        links="newline",
        plain=False,
        strip_html=False,
        base_url="",
        md_links=False,
        link_func=None,
        table_tag="table",
        checklist=True,
    ):
        # Disable all the HTML renderer's messing around:
        super().__init__(escape=False, allow_harmful_protocols=True)

        self.md_links = md_links
        self.link_func = link_func
        if base_url is None:
            base_url = ""
        if len(base_url) > 0 and base_url[-1] == "/":
            base_url = base_url[:-1]
        self.base_url = base_url
        self.plain = plain
        self.strip_html = strip_html
        self.ascii = ascii_table
        if indent is None:
            self.indent = "  "
        else:
            self.indent = indent
        if code_tag is None:
            code_tag = ""
        self.code_tag = code_tag
        if img_tag is None:
            img_tag = ""
        self.img_tag = " " + img_tag
        if table_tag is None:
            table_tag = "table"
        self.table_tag = table_tag
        self.checklist = checklist
        # Tables
        self.unitable = None
        self.table_cols_align = []  # List of column alignments: ["l", "r", "c"]
        # Footnote links
        self.links = links
        if self.links in ["paragraph", "at-end", "copy"]:
            self.footnotes_enabled = True
        else:
            self.footnotes_enabled = False
        self.footnote_num = 0  # The number of the last footnote is stored here
        self.footnotes = (
            []
        )  # ["https://example.com/", ...] - footnotes per paragraph/document stored here
        self.footnote_texts = (
            []
        )  # ["link text", ...] - used for links "copy" mode, when link text also needs to be stored

    def _gem_link(self, link, text=None):
        # Links are handled in post processing, these control characters
        # are just used to denote paragraph start and end. They were picked
        # because they will never be typed in normal text editing.

        link = link.strip()
        if link.startswith("/"):
            link = self.base_url + link

        if text is None:
            return LINK_DELIM + "=> " + link.strip() + LINK_DELIM
        return LINK_DELIM + "=> " + link.strip() + " " + text.strip() + LINK_DELIM

    def _add_footnote(self, link, text):
        self.footnote_num += 1
        self.footnotes.append(link)

        if self.links == "copy":
            self.footnote_texts.append(text)
            return text
        return text + "[" + str(self.footnote_num) + "]"

    def _render_footnotes(self):
        if not self.footnotes_enabled:
            return ""
        if self.footnotes == []:
            return ""

        ret = ""
        length = len(self.footnotes)
        if self.links == "copy":
            for i, url in enumerate(self.footnotes):
                # Calculate the relative footnote number - there could be five footnotes
                # for this paragraph, but now 10 in total.

                # Example footnote, in a client view:
                # copied link text
                # Actual footnote output:
                # => gemini://gus.guru/ copied link text
                ret += self._gem_link(url, self.footnote_texts[i])
        else:
            for i, url in enumerate(self.footnotes):
                # Example footnote, in a client view:
                # 10: gemini://gus.guru/
                # Actual footnote output:
                # => gemini://gus.guru/ 10: gemini://gus.guru/
                ret += self._gem_link(
                    url, str((self.footnote_num - length) + 1 + i) + ": " + url.strip()
                )

        return ret

    # Inline elements

    def text(self, text):
        return text

    def link(self, link, text=None, title=None):
        # title is ignored because it doesn't apply to Gemini

        if link.startswith("#"):
            # Invalid for Gemini
            if text is None:
                return link
            return text
        if self.links == "off":
            # Don't link, just leave the text as it was written
            if text is None:
                return link
            return text

        if self.md_links and "//" not in link:
            # Relative link, and md -> gmi conversion is enabled
            if link.endswith(".md"):
                link = link[:-2] + "gmi"
            elif ".md#" in link:
                index = link.index(".md#")
                link = link[: index + 1] + "gmi"

        if callable(self.link_func):
            # A custom function to treat links is provided
            link = self.link_func(link)

        if self.footnotes_enabled:
            if text is None or text.strip() == "":
                # Insert the link inline, but with a footnote too
                return self._add_footnote(link, link)
            else:
                return self._add_footnote(link, text)

        return self._gem_link(link, text)

    def image(self, src, alt="", title=None):
        """Turn images into regular Gemini links."""

        if alt is None:
            alt = ""

        if self.links == "off":
            if alt == "":
                return src
            return alt

        # Images shouldn't need footnotes, so it's just added as a link no matter what

        return self._gem_link(src, alt.strip() + self.img_tag)

    def emphasis(self, text):
        if self.plain:
            return text
        return "*" + text + "*"

    def strong(self, text):
        if self.plain:
            return text
        return "**" + text + "**"

    def codespan(self, text):
        if self.plain:
            return text
        return "`" + text + "`"

    def linebreak(self):
        # return "<LB>"
        return LINEBREAK

    def newline(self):
        # return "<NL>"
        return ""

    def inline_html(self, html):
        if self.plain or self.strip_html:
            return ""
        return html

    # Block level elements

    def paragraph(self, text):
        # Paragraphs are handled in post processing, these control characters
        # are just used to denote paragraph start and end. They were picked
        # because they will never be typed in normal text editing.

        if (
            self.links == "paragraph"
            and self.footnotes_enabled
            and text.count("\n") <= 1
            and len(self.footnotes) == 1
            and text.rstrip().endswith("[" + str(self.footnote_num) + "]")
        ):
            # The whole paragraph is just one line, just the link
            # So there shouldn't be a footnote
            ret = (
                PARAGRAPH_DELIM
                + self._gem_link(
                    self.footnotes[0],
                    # Remove the footnote part from the text, the [X] at the end
                    text.rstrip()[: -(len(str(self.footnote_num)) + 2)],
                )
                + PARAGRAPH_DELIM
            )
            # Remove footnote from list
            self.footnotes.pop()
            self.footnote_num -= 1
            self.footnotes = []  # Reset them for the next paragraph
            return ret

        if (
            self.links == "copy"
            and len(self.footnotes) == 1
            and self.footnote_texts[0] == text
        ):
            # The whole paragraph is just one big link, so it should just be added as a link
            ret = (
                PARAGRAPH_DELIM
                + self._gem_link(self.footnotes[0], text)
                + PARAGRAPH_DELIM
            )
            self.footnotes = []
            self.footnote_texts = []
            return ret

        text += self._end_of_paragraph()

        return PARAGRAPH_DELIM + text + PARAGRAPH_DELIM

    def _end_of_paragraph(self):
        # Process footnotes if it should
        if self.links in ["paragraph", "copy"] and len(self.footnotes) > 0:
            ret = PARAGRAPH_DELIM + self._render_footnotes() + PARAGRAPH_DELIM
            self.footnotes = []
            self.footnote_texts = []  # For self.links == "copy"
            return ret

        return ""

    def heading(self, text, level):
        return self._end_of_paragraph() + "#" * level + " " + text + NEWLINE * 2

    def thematic_break(self):
        """80 column split using hyphens."""

        return "-" * 80 + NEWLINE * 2

    def block_text(self, text):
        # Idk what this is, it's not defined in the CommonMark spec,
        # and the HTML renderer also just returns text
        return text + NEWLINE

    def block_code(self, code, info=None):
        # Gemini doesn't support code block infos, but it doesn't matter
        # Adding them might make this more compatible
        start = "```" + self.code_tag + NEWLINE
        if not info is None:
            start = "```" + info + NEWLINE

        if code.endswith("\n"):
            code = code[:-1]
        if code.endswith("\n"):
            code = code[:-1]
        return start + code + NEWLINE + "```" + NEWLINE * 2

    def block_quote(self, text):
        """Add a quote mark to the beginning of each line."""

        lines = (
            text.replace(PARAGRAPH_DELIM, "\n")
            .replace(LINEBREAK, "\n")
            .strip()
            .splitlines()
        )

        ret = ""
        for i, line in enumerate(lines):
            line = line.strip()
            link = line and line[0] == LINK_DELIM
            ret += LINEBREAK if link else "> "
            ret += line
            if i < len(lines) - 1:
                # Linebreak used to prevent removal later
                ret += LINEBREAK

        return PARAGRAPH_DELIM + ret + PARAGRAPH_DELIM

    def block_html(self, html):
        if self.strip_html:
            return NEWLINE
        return self.block_code(html, "html")

    def block_error(self, html):
        return self.block_code(html, "html")

    def list_item(self, text, level):
        # It is necessary to split the text on newline, since markdown
        # allows for list items to be split across multiple lines.
        # We need to strip whitespace from these items and add it ourselves,
        # since the text doesn't guarantee any particular formatting for
        # these items.
        new_text = ""
        last_offset = 0
        in_fence = False
        text = text.replace(PARAGRAPH_DELIM, PARAGRAPH_DELIM + "\r\n")
        for item in text.splitlines():
            was_in_fence = in_fence
            m = FENCE_EXPR.match(item)
            if m:
                this_offset = len(m.groups()[0])
                in_fence = (
                    not in_fence if m and this_offset == last_offset else in_fence
                )
                if in_fence and not was_in_fence:
                    last_offset = this_offset
            if in_fence:
                new_text += LINEBREAK + item
            else:
                if was_in_fence:
                    new_text += LINEBREAK + item + LINEBREAK
                else:
                    if new_text:
                        new_text += " " + item.lstrip()
                    else:
                        new_text = item
        return new_text + NEWLINE

    def list(self, text, ordered, level, start=None):
        """Gemini only defines single-level unordered lists.

        This uses indenting to do sub-levels.
        Ordered list items just use 1. 2. etc, as plain text.
        """

        # First level of list means `level = 1`

        if start is None:
            start = 1
        # Any empty lines should have been consumed by list_item,
        # so no need to check for empty lines here.
        items = text.splitlines()
        ret_items = []

        if ordered:
            # Recreate the ordered list for each item, then return it
            # This just returns a plain text list.
            for i, item in enumerate(items):
                if item.startswith(self.indent):
                    # It's an already processed sub-level item, don't do anything.
                    # This kind of check is needed because mistune passes nested list items
                    # to the renderer multiple times. Once as N-level list, and then once again
                    # as a part of a flattened level 1 list.
                    # This check prevents the item from being processed again when it appears
                    # in that second list.
                    ret_items.append(item)
                else:
                    ret_items.append(
                        self.indent * (level - 1) + str(i + start) + ". " + item.strip()
                    )

        else:
            # Return an unordered list using the official Gemini list character: *
            # Indenting is used for sub-levels
            for item in items:
                if item.startswith(self.indent):
                    # See the comment above for why this check is required.
                    ret_items.append(item)
                else:
                    ret_items.append(self.indent * (level - 1) + "* " + item.strip())

        return (
            PARAGRAPH_DELIM
            + LINEBREAK.join(ret_items)
            + self._end_of_paragraph()
            + PARAGRAPH_DELIM
        )

    # Elements that rely on plugins:

    # Tables
    # Most of the funcs just return the text unchanged because the actual text processing
    # is done at the end, using the UniTable class.

    def _init_table(self):
        self.unitable = UniTable()
        if self.ascii:
            self.unitable.set_style("default")
        else:
            self.unitable.set_style("light")

    def table(self, text):
        # Called at the end I think, once all the table elements
        # have been processed
        # Put the table in a preprocessed block
        return (
            "```"
            + self.table_tag
            + NEWLINE
            + self.unitable.draw()
            + NEWLINE
            + "```"
            + NEWLINE
        )

    def table_head(self, text):
        self._init_table()
        # The table_cell func splits each column using newlines
        try:
            self.unitable.header(
                text.split("\n")[:-1]  # \n is used to delimit cells internally
            )
        except ArraySizeError:
            # raise Exception("Malformed table")
            pass
        # Set and clear the alignment data, now that the number of columns should be known
        self.unitable.set_cols_align(self.table_cols_align)
        self.unitable.set_cols_valign(["m"] * len(self.table_cols_align))
        self.table_cols_align = []
        return text

    def table_body(self, text):
        return ""

    def table_row(self, text):
        try:
            self.unitable.add_row(
                # The table_cell func splits each column using newlines
                text.split("\n")[:-1]  # \n is used to delimit cells internally
            )
        except ArraySizeError:
            # raise Exception("Malformed table")
            pass
        self.table_cols_align = []
        # The text processing is done in other funcs
        return text

    def table_cell(self, text, align=None, is_head=False):
        if align in ["left", "right", "center"]:
            self.table_cols_align.append(align[0])  # l, r, or c
        else:
            # If align is None or something unknown happened
            self.table_cols_align.append("l")
        return (
            text.strip() + "\n"
        )  # \n is used to separate cells from each other in other funcs

    # Task list / check list support
    # https://github.com/makeworld-the-better-one/md2gemini/issues/19

    def task_list_item(self, text, level, checked):
        if self.checklist:
            if checked:
                symbol = "\U0001f5f9"  # BALLOT BOX WITH BOLD CHECK
            else:
                symbol = "\u2610"  # BALLOT BOX
        else:
            if checked:
                symbol = "[x]"
            else:
                symbol = "[ ]"
        return self.list_item(symbol + " " + text, level)

    # Strikethough can't be supported
    # Footnotes aren't supported right now



from .util import normalize
from md2gemini import md2gemini


def f(md):
    return normalize(md2gemini(md))


def test_single_line_quote():
    md = "> test quote"
    gem = md
    assert f(md) == gem


def test_multiline_quote():
    md = "> test quote\n>test again\n>  third time"
    gem = "> test quote\n> test again\n> third time"
    assert f(md) == gem


def test_quote_with_hard_linebreaks():
    # https://github.com/makeworld-the-better-one/md2gemini/issues/32

    md = """
> What is a good man?  
> A teacher of a bad man.  
> What is a bad man?  
> A good man's charge.
"""

    gem = """
> What is a good man?
> A teacher of a bad man.
> What is a bad man?
> A good man's charge.
""".strip()

    assert f(md) == gem


def test_quote_with_text_after():
    # https://github.com/makeworld-the-better-one/md2gemini/issues/31

    md = """
> some quote

More text
    """

    gem = """
> some quote

More text
    """.strip()

    assert f(md) == gem



# -*- coding: utf-8 -*-

# The below code was NOT written by me.
# It was taken from https://github.com/fariello/texttable/blob/master/unitable.py
# which is a fork by @fariello of the original texttable python package by @foutaise
# According to the LICENSE file in that repo, this code is under the MIT license,
# meaning it is legal for me to reuse it here.
# Here is the LICENSE text, that applies to this file ONLY:
# The MIT License (MIT)

# Copyright (c) 2019 Gerome Fournier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# After this line, the original copied file begins:

# texttable - module for creating simple ASCII tables
# Copyright (C) 2003-2019 Gerome Fournier <jef(at)foutaise.org>

"""module for creating simple ASCII tables


Example:

    table = UniTable()
    table.set_cols_align(["l", "r", "c"])
    table.set_cols_valign(["t", "m", "b"])
    table.add_rows([["Name", "Age", "Nickname"],
                    ["Mr\\nXavier\\nHuon", 32, "Xav'"],
                    ["Mr\\nBaptiste\\nClement", 1, "Baby"],
                    ["Mme\\nLouise\\nBourgeau", 28, "Lou\\n\\nLoue"]])
    print table.draw() + "\\n"

    table = UniTable()
    table.set_deco(UniTable.HEADER)
    table.set_cols_dtype(['t',  # text
                          'f',  # float (decimal)
                          'e',  # float (exponent)
                          'i',  # integer
                          'a']) # automatic
    table.set_cols_align(["l", "r", "r", "r", "l"])
    table.add_rows([["text",    "float", "exp", "int", "auto"],
                    ["abcd",    "67",    654,   89,    128.001],
                    ["efghijk", 67.5434, .654,  89.6,  12800000000000000000000.00023],
                    ["lmn",     5e-78,   5e-78, 89.4,  .000000000000128],
                    ["opqrstu", .023,    5e+78, 92.,   12800000000000000000000]])
    print table.draw()

Result:

    +----------+-----+----------+
    |   Name   | Age | Nickname |
    +==========+=====+==========+
    | Mr       |     |          |
    | Xavier   |  32 |          |
    | Huon     |     |   Xav'   |
    +----------+-----+----------+
    | Mr       |     |          |
    | Baptiste |   1 |          |
    | Clement  |     |   Baby   |
    +----------+-----+----------+
    | Mme      |     |   Lou    |
    | Louise   |  28 |          |
    | Bourgeau |     |   Loue   |
    +----------+-----+----------+

    text   float       exp      int     auto
    ===========================================
    abcd   67.000   6.540e+02   89    128.001
    efgh   67.543   6.540e-01   90    1.280e+22
    ijkl   0.000    5.000e-78   89    0.000
    mnop   0.023    5.000e+78   92    1.280e+22
"""

from __future__ import division

# @makeworld-the-better-one removed the first underscore of each of those variables
# to help make sure packages don't get messed up

_all__ = ["UniTable", "ArraySizeError"]

_author__ = 'Gerome Fournier <jef(at)foutaise.org>'
_license__ = 'MIT'
_version__ = '1.6.2'
_credits__ = """\
Jeff Kowalczyk:
    - textwrap improved import
    - comment concerning header output

Anonymous:
    - add_rows method, for adding rows in one go

Sergey Simonenko:
    - redefined len() function to deal with non-ASCII characters

Roger Lew:
    - columns datatype specifications

Brian Peterson:
    - better handling of unicode errors

Frank Sachsenheim:
    - add Python 2/3-compatibility

Maximilian Hils:
    - fix minor bug for Python 3 compatibility

frinkelpi:
    - preserve empty lines

gfariello:
    - Added unicode box border options, made errors more informative, corrected typos
"""

import sys
import unicodedata

# define a text wrapping function to wrap some text
# to a specific width:
# - use cjkwrap if available (better CJK support)
# - fallback to textwrap otherwise
try:
    import cjkwrap
    def textwrapper(txt, width):
        return cjkwrap.wrap(txt, width)
except ImportError:
    try:
        import textwrap
        def textwrapper(txt, width):
            return textwrap.wrap(txt, width)
    except ImportError:
        sys.stderr.write("Can't import textwrap module!\n")
        raise

# define a function to calculate the rendering width of a unicode character
# - use wcwidth if available
# - fallback to unicodedata information otherwise
try:
    import wcwidth
    def uchar_width(c):
        """Return the rendering width of a unicode character
        """
        return max(0, wcwidth.wcwidth(c))
except ImportError:
    def uchar_width(c):
        """Return the rendering width of a unicode character
        """
        if unicodedata.east_asian_width(c) in 'WF':
            return 2
        elif unicodedata.combining(c):
            return 0
        else:
            return 1

from functools import reduce

if sys.version_info >= (3, 0):
    unicode_type = str
    bytes_type = bytes
    str_class = str
else:
    unicode_type = unicode
    bytes_type = str
    str_class = basestring
    pass

def obj2unicode(obj):
    """Return a unicode representation of a python object"""
    if isinstance(obj, unicode_type):
        return obj
    elif isinstance(obj, bytes_type):
        try:
            return unicode_type(obj, 'utf-8')
        except UnicodeDecodeError as strerror:
            sys.stderr.write("UnicodeDecodeError exception for string '%s': %s\n" % (obj, strerror))
            return unicode_type(obj, 'utf-8', 'replace')
    else:
        return unicode_type(obj)


def len(iterable):
    """Redefining len here so it will be able to work with non-ASCII characters    """
    if isinstance(iterable, bytes_type) or isinstance(iterable, unicode_type):
        return sum([uchar_width(c) for c in obj2unicode(iterable)])
    else:
        return iterable.__len__()


class ArraySizeError(Exception):
    """Exception raised when specified rows don't fit the required size"""
    def __init__(self, msg):
        self.msg = msg
        Exception.__init__(self, msg, '')
    def __str__(self):
        return self.msg

class FallbackToText(Exception):
    """Used for failed conversion to float"""
    pass

2
class UTableBaseClass:
    """
    This is the base class for all UTable classes. It should probably not be instantiated directly.
    """
    def __init__(self,parent=None):
        self.parent = parent
        self.needs_recalc = True
        pass
    def _chktype(self,what,clss,val,none_ok=False):
        if none_ok and val is None:
            return val
        if not isinstance(val, clss):
            raise ValueError("ERROR: %s must be an instance of %s, '%s' is a %s." %(what,clss.__name__,val,type(val).__name__))
        return val
    def set_parent(self,val,clss):
        self._parent = self._chktype('parent',clss,val,True)
        pass
    @property
    def parent(self):
        return self._parent
    @parent.setter
    def parent(self,val):
        self.set_parent(val)
        pass
    @property
    def needs_recalc(self):
        return self._needs_recalc
    @needs_recalc.setter
    def needs_recalc(self,val):
        self._chktype('needs_recalc',bool,val)
        if self.parent is not None:
            self.parent.needs_recalc = val
            pass
        self._needs_recalc = val
        pass
    @property
    def width(self):
        return self._width;
    @width.setter
    def width(self,val):
        self._width = self._chktype('width',int,val,True)
    @property
    def witdth(self): return self._width;
    @width.setter
    def height(self,val):
        self._height = self._chktype('height',int,val,True)

class UTable:
    def __init__(self,rows=[],parent=None):
        super().__init__(self,parent)
        self._min_width = None
        self._max_width = None
        self._height = None
        self._rows = []
        self._cols = []
        pass
    pass

class URow:
    def __init__(self,table,cells=[]):
        super().__init__(parent)
        pass
    pass

class UCol:
    def __init__(self,table,cells=[]):
        super().__init__(parent)
        pass
    pass

class Cell:
    def __init__(self,value="",row=None,col=None):
        self._recalc = True
        self.value = value
        self.row = row
        self.col = column
        pass
    @property
    def value(self):
        return self._value
    @value.setter
    def value(self,val):
        if val == self._value:
            return
        self._recalc = True
        if isinstance(val,str_class):
            self._value = val.splitlines()
        else:
            self._value = [val]
            pass
        pass
    @property
    def row(self):
        return self._row
    @row.setter
    def row(self,val):
        if val == self._row:
            return
        if val is None or isinstance(val,int):
            self._row = val
            self._recalc = True
            pass
        raise ValueError("ERROR: row must be an integer. Received '%s' which is a %s." %(val,type(val).__name__))
    @property
    def col(self):
        return self._col
    @col.setter
    def col(self,val):
        if val == self._col:
            return
        if val is None or isinstance(val,int):
            self._col = val
            self._recalc = True
            pass
        raise ValueError("ERROR: col must be an integer. Received '%s' which is a %s." %(val,type(val).__name__))
    @property
    def width(self):
        return self._width

class UniTable:
    BORDER = 1
    HEADER = 1 << 1
    HLINES = 1 << 2
    VLINES = 1 << 3
    # --- gfariello -- Start -- Added to support new styles.
    TOP = 0
    MIDDLE  = 1
    BOTTOM = 2
    STYLES = {
        "bold":         "━┃┏┓┗┛┣┫┳┻╋━┣┫╋",
        "default":      "-|+=",
        "double":       "═║╔╗╚╝╠╣╦╩╬═╠╣╬",
        "very_light":   "─│┌┐└┘├┤┬┴┼─├┤┼",
        "light":        "─│┌┐└┘├┤┬┴┼═╞╡╪",
        "round":        "─│╭╮╰╯├┤┬┴┼─├┤┼",
        "round2":       "─│╭╮╰╯├┤┬┴┼═╞╡╪",
        "simple":       "-|+-",
        }
    STYLE_MAPPER = {
        "heavy": {
            "---w": " ",
            "--e-": " ",
            "--ew": "━",
            "-s--": " ",
            "-s-w": "┓",
            "-se-": "┏",
            "-sew": "┳",
            "n---": " ",
            "n--w": "┛",
            "n-e-": "┗",
            "n-ew": "┻",
            "ns--": "┃",
            "ns-w": "┫",
            "nse-": "┣",
            "nsew": "╋",
        },
        "light": {
            "---w": " ",
            "--e-": " ",
            "--ew": "-",
            "-s--": " ",
            "-s-w": "┐",
            "-se-": "┌",
            "-sew": "┬",
            "n---": " ",
            "n--w": "┘",
            "n-e-": "└",
            "n-ew": "┴",
            "ns--": "│",
            "ns-w": "┤",
            "nse-": "├",
            "nsew": "┼",
        },
        "round": {
            "---w": " ",
            "--e-": " ",
            "--ew": "-",
            "-s--": " ",
            "-s-w": "╮",
            "-se-": "╭",
            "-sew": "┬",
            "n---": " ",
            "n--w": "╯",
            "n-e-": "╰",
            "n-ew": "┴",
            "ns--": "│",
            "ns-w": "┤",
            "nse-": "├",
            "nsew": "┼",
        },
        "double": {
            "---w": " ",
            "--e-": " ",
            "--ew": "═",
            "-s--": " ",
            "-s-w": "╗",
            "-se-": "╔",
            "-sew": "╦",
            "n---": " ",
            "n--w": "╝",
            "n-e-": "╚",
            "n-ew": "╩",
            "ns--": "║",
            "ns-w": "╣",
            "nse-": "╠",
            "nsew": "╬",
        },
        "heavy:light": {
            "---w:--e-": "╾",
            "---w:-s--": "┑",
            "---w:-se-": "┲",
            "---w:n---": "┙",
            "---w:n-e-": "┺",
            "---w:ns--": "┥",
            "---w:nse-": "┽",
            "--e-:---w": "╼",
            "--e-:-s--": "┍",
            "--e-:-s-w": "┮",
            "--e-:n---": "┙",
            "--e-:n--w": "┶",
            "--e-:ns--": "┝",
            "--e-:ns-w": "┾",
            "--ew:-s--": "┰",
            "--ew:n---": "┸",
            "--ew:ns--": "┿",
            "-s--:---w": "┒",
            "-s--:--e-": "┎",
            "-s--:--ew": "┰",
            "-s--:n---": "╽",
            "-s--:n--w": "┧",
            "-s--:n-e-": "┟",
            "-s--:n-ew": "╁",
            "-s-w:--e-": "┱",
            "-s-w:n---": "┧",
            "-s-w:n-e-": "╅",
            "-se-:---w": "┲",
            "-se-:n---": "┢",
            "-se-:n--w": "╆",
            "-sew:n---": "╈",
            "n---:---w": "┖",
            "n---:--e-": "┚",
            "n---:--ew": "┸",
            "n---:-s--": "╿",
            "n---:-s-w": "┦",
            "n---:-se-": "┞",
            "n---:-sew": "╀",
            "n--w:--e-": "┹",
            "n--w:-s--": "┩",
            "n--w:-se-": "╃",
            "n-e-:---w": "┺",
            "n-e-:-s--": "┡",
            "n-e-:-s-w": "╄",
            "n-ew:-s--": "╇",
            "ns--:---w": "┨",
            "ns--:--e-": "┠",
            "ns--:--ew": "╂",
            "ns-w:--e-": "╉",
            "nse-:---w": "╊",
            }
            }
    # --- gfariello -- End -- Added to support new styles.

    # --- gfariello -- Start -- Added init with table def.
    # NOTE: See below about backward compatability
    def __init__(self, rows=None, max_width=80):
    # --- gfariello -- End -- Added init with table def.
        """Constructor

        - max_width is an integer, specifying the maximum width of the table
        - if set to 0, size is unlimited, therefore cells won't be wrapped
        """
        self._has_border = True
        self._has_header = True
        self._has_hline_between_headers = True
        self._has_hline_header_2_cell = True
        self._has_hline_between_cells = True
        self._has_vline_between_headers = True
        self._has_vline_header_2_cell = True
        self._has_vline_between_cells = True
        self.set_max_width(max_width)
        self._precision = 3

        self._deco = UniTable.VLINES | UniTable.HLINES | UniTable.BORDER | \
            UniTable.HEADER
        self.set_style("default")
        self._pad = 1
        self.reset()
        # --- gfariello -- Start -- Added to support rows arg (i.e., adding
        # entire table definition in initilization). NOTE: This changed the
        # order (max_width is now one arg later) and therefore has a chance of
        # breaking older code that called UniTable(50) but not
        # UniTable(max_width=50). It felt less intuitive to have the rows
        # definition after the max_width, but if backwards compatibility is
        # more important, just swap the order of rows and max_width.
        if rows is not None:
            self.add_rows(rows)
            pass
        # --- gfariello -- End -- Added to support rows arg.
        pass
    @property
    def has_border(self):
        return self._has_border
    @has_border.setter
    def has_border(self,value):
        self._has_border = value
        return value

    @property
    def has_header(self):
        return self._has_header
    @has_header.setter
    def has_header(self,value):
        self._has_header = value
        return value

    def reset(self):
        """Reset the instance

        - reset rows and header
        """

        self._hline_string = None
        self._row_size = None
        self._header = []
        self._rows = []
        return self

    def set_max_width(self, max_width):
        """Set the maximum width of the table

        - max_width is an integer, specifying the maximum width of the table
        - if set to 0, size is unlimited, therefore cells won't be wrapped
        """
        self._max_width = max_width if max_width > 0 else False
        return self

    def set_style(self, style="light"):
        """Set the characters used to draw lines between rows and columns to one of four box types:

        "light": Use unicode light box borders (─│┌┐└┘├┤┬┴┼)
        "bold":  Use unicode bold box borders (━┃┏┓┗┛┣┫┳┻╋)
        "double": Use unicode double box borders (═║╔╗╚╝╠╣╦╩╬)

        Default if none provided is "light"
        """
        if style in UniTable.STYLES:
            return self.set_chars(UniTable.STYLES[style])
        raise ValueError("style must be one of '%s' not '%s'" %("','".join(sorted(UniTable.STYLES.keys())),style))

    def _set_chars(self, array):
        """Set the characters used to draw lines between rows and columns in the following format:
        [
          ew,    # The character connecting east and west to use for a horizantal line (e.g. "-" or "─" )
          ns,    # The character connecting north and south to use for a vertical line (e.g. "|" or "|" )
          se,    # The character connecting south and east to use for the top- and left-most corner (e.g. "+", or "┌")
          sw,    # The character connecting south and west to use for the top- and right-most corner (e.g. "+" or "┐")
          ne,    # The character connecting north and east to use for the bottom- and left-most corner (e.g. "+" or "└")
          nw,    # The character connecting north and west to use for the bottom- and right-most corner (e.g. "+" or "┘")
          nse,   # The character connecting north, south, and east (e.g., "+" or "┤")
          nsw,   # The character connecting north, south, and west (e.g., "+" or "├")
          sew,   # The character connecting south, east, and west (e.g., "+" or "┬")
          new,   # The character connecting north, east, and west (e.g., "+" or "┴")
          nsew,  # The character connecting north, south, east, and west (e.g., "+" or "┴")
          hew,   # The character connecting east and west to use for a line separating headers (e.g. "=" or "═" )
          hnse,  # The character connecting north, south and east to use for a line separating headers (e.g. "+" or "╞" )
          hnsw,  # The character connecting north, south, and west to use for a line separating headers (e.g. "+" or "╡" )
          hnsew, # The character connecting north, south, east and west to use for a line separating headers (e.g. "+" or "╪" )
        ]
        For legacy default it would be "-|+++++++++=+++"
        """
        if len(array) != 15:
            raise ArraySizeError("string/array should contain 15 characters not %d as in '%s'" %(len(array),array))
        (
            self._char_ew,
            self._char_ns,
            self._char_se,
            self._char_sw,
            self._char_ne,
            self._char_nw,
            self._char_nse,
            self._char_nsw,
            self._char_sew,
            self._char_new,
            self._char_nsew,
            self._char_hew,
            self._char_hnse,
            self._char_hnsw,
            self._char_hnsew,
        ) = array
        return self

    def set_chars(self, array):
        """Set the characters used to draw lines between rows and columns

        - the array should contain 4 fields:

            [horizontal, vertical, corner, header]

        - default is set to:

            ['-', '|', '+', '=']
        """
        if len(array) == 15:
            return self._set_chars(array)
        if len(array) != 4:
            raise ArraySizeError("string/array should contain either 4 or 15 characters not %d as in '%s'" %(len(array),array))
        (hor,ver,cor,hea) = array
        return self._set_chars([hor,ver,cor,cor,cor,cor,cor,cor,cor,cor,cor,hea,cor,cor,cor])

    def set_deco(self, deco):
        """Set the table decoration

        - 'deco' can be a combinasion of:

            UniTable.BORDER: Border around the table
            UniTable.HEADER: Horizontal line below the header
            UniTable.HLINES: Horizontal lines between rows
            UniTable.VLINES: Vertical lines between columns

           All of them are enabled by default

        - example:

            UniTable.BORDER | UniTable.HEADER
        """

        self._deco = deco
        return self

    def set_header_align(self, array):
        """Set the desired header alignment

        - the elements of the array should be either "l", "c" or "r":

            * "l": column flushed left
            * "c": column centered
            * "r": column flushed right
        """

        self._check_row_size(array)
        self._header_align = array
        return self

    def set_cols_align(self, array):
        """Set the desired columns alignment

        - the elements of the array should be either "l", "c" or "r":

            * "l": column flushed left
            * "c": column centered
            * "r": column flushed right
        """

        self._check_row_size(array)
        self._align = array
        return self

    def set_cols_valign(self, array):
        """Set the desired columns vertical alignment

        - the elements of the array should be either "t", "m" or "b":

            * "t": column aligned on the top of the cell
            * "m": column aligned on the middle of the cell
            * "b": column aligned on the bottom of the cell
        """

        self._check_row_size(array)
        self._valign = array
        return self

    def set_cols_dtype(self, array):
        """Set the desired columns datatype for the cols.

        - the elements of the array should be either a callable or any of
          "a", "t", "f", "e" or "i":

            * "a": automatic (try to use the most appropriate datatype)
            * "t": treat as text
            * "f": treat as float in decimal format
            * "e": treat as float in exponential format
            * "i": treat as int
            * a callable: should return formatted string for any value given

        - by default, automatic datatyping is used for each column
        """

        self._check_row_size(array)
        self._dtype = array
        return self

    def set_cols_width(self, array):
        """Set the desired columns width

        - the elements of the array should be integers, specifying the
          width of each column. For example:

                [10, 20, 5]
        """

        self._check_row_size(array)
        try:
            array = list(map(int, array))
            if reduce(min, array) <= 0:
                raise ValueError
        except ValueError:
            sys.stderr.write("Wrong argument in column width specification\n")
            raise
        self._width = array
        return self

    def set_precision(self, width):
        """Set the desired precision for float/exponential formats

        - width must be an integer >= 0

        - default value is set to 3
        """

        if not type(width) is int or width < 0:
            raise ValueError('width must be an integer greater then 0')
        self._precision = width
        return self

    def set_padding(self, amount):
        """Set the amount of spaces to pad cells (right and left, we don't do top bottom padding)

        - width must be an integer >= 0

        - default value is set to 1
        """
        if not type(amount) is int or amount < 0:
            raise ValueError('padding must be an integer greater then 0')
        self._pad = amount
        return self

    def header(self, array):
        """Specify the header of the table
        """

        self._check_row_size(array)
        self._header = list(map(obj2unicode, array))
        return self

    def add_row(self, array):
        """Add a row in the rows stack

        - cells can contain newlines and tabs
        """

        self._check_row_size(array)

        if not hasattr(self, "_dtype"):
            self._dtype = ["a"] * self._row_size

        cells = []
        for i, x in enumerate(array):
            cells.append(self._str(i, x))
        self._rows.append(cells)
        return self

    def add_rows(self, rows, header=True):
        """Add several rows in the rows stack

        - The 'rows' argument can be either an iterator returning arrays,
          or a by-dimensional array
        - 'header' specifies if the first row should be used as the header
          of the table
        """

        # nb: don't use 'iter' on by-dimensional arrays, to get a
        #     usable code for python 2.1
        if header:
            if hasattr(rows, '__iter__') and hasattr(rows, 'next'):
                self.header(rows.next())
            else:
                self.header(rows[0])
                rows = rows[1:]
        for row in rows:
            self.add_row(row)
        return self

    def draw(self):
        """Draw the table

        - the table is returned as a whole string
        """

        if not self._header and not self._rows:
            return
        self._compute_cols_width()
        self._check_align()
        out = ""
        if self.has_border:
            out += self._hline(location=UniTable.TOP)
        if self._header:
            out += self._draw_line(self._header, isheader=True)
            if self.has_header:
                out += self._hline_header(location=UniTable.MIDDLE)
                pass
            pass
        num = 0
        length = len(self._rows)
        for row in self._rows:
            num += 1
            out += self._draw_line(row)
            if self.has_hlines() and num < length:
                out += self._hline(location=UniTable.MIDDLE)
        if self._has_border:
            out += self._hline(location=UniTable.BOTTOM)
        return out[:-1]

    @classmethod
    def _to_float(cls, x):
        if x is None:
            raise FallbackToText()
        try:
            return float(x)
        except (TypeError, ValueError):
            raise FallbackToText()

    @classmethod
    def _fmt_int(cls, x, **kw):
        """Integer formatting class-method.

        - x will be float-converted and then used.
        """
        return str(int(round(cls._to_float(x))))

    @classmethod
    def _fmt_float(cls, x, **kw):
        """Float formatting class-method.

        - x parameter is ignored. Instead kw-argument f being x float-converted
          will be used.

        - precision will be taken from `n` kw-argument.
        """
        n = kw.get('n')
        return '%.*f' % (n, cls._to_float(x))

    @classmethod
    def _fmt_exp(cls, x, **kw):
        """Exponential formatting class-method.

        - x parameter is ignored. Instead kw-argument f being x float-converted
          will be used.

        - precision will be taken from `n` kw-argument.
        """
        n = kw.get('n')
        return '%.*e' % (n, cls._to_float(x))

    @classmethod
    def _fmt_text(cls, x, **kw):
        """String formatting class-method."""
        return obj2unicode(x)

    @classmethod
    def _fmt_auto(cls, x, **kw):
        """auto formatting class-method."""
        f = cls._to_float(x)
        if abs(f) > 1e8:
            fn = cls._fmt_exp
        elif f != f:  # NaN
            fn = cls._fmt_text
        elif f - round(f) == 0:
            fn = cls._fmt_int
        else:
            fn = cls._fmt_float
        return fn(x, **kw)

    def _str(self, i, x):
        """Handles string formatting of cell data

            i - index of the cell datatype in self._dtype
            x - cell data to format
        """
        FMT = {
            'a':self._fmt_auto,
            'i':self._fmt_int,
            'f':self._fmt_float,
            'e':self._fmt_exp,
            't':self._fmt_text,
            }

        n = self._precision
        dtype = self._dtype[i]
        try:
            if callable(dtype):
                return dtype(x)
            else:
                return FMT[dtype](x, n=n)
        except FallbackToText:
            return self._fmt_text(x)

    def _check_row_size(self, array):
        """Check that the specified array fits the previous rows size
        """

        if not self._row_size:
            self._row_size = len(array)
        elif self._row_size != len(array):
            raise ArraySizeError("array should contain %d elements not %s (array=%s)" \
                %(self._row_size,len(array),array))

    def has_vlines(self):
        """Return a boolean, if vlines are required or not
        """

        return self._deco & UniTable.VLINES > 0

    def has_hlines(self):
        """Return a boolean, if hlines are required or not
        """

        return self._deco & UniTable.HLINES > 0

    def _hline_header(self,location=MIDDLE):
        """Print header's horizontal line
        """

        return self._build_hline(is_header=True,location=location)

    def _hline(self,location):
        """Print an horizontal line
        """
        # if not self._hline_string:
        #   self._hline_string = self._build_hline(location)
        # return self._hline_string
        return self._build_hline(is_header=False,location=location)

    def _build_hline(self, is_header=False, location=MIDDLE):
        """Return a string used to separated rows or separate header from
        rows
        """
        horiz_char = self._char_hew if is_header else self._char_ew
        if UniTable.TOP == location:
            left, mid, right = self._char_se, self._char_sew, self._char_sw
        elif UniTable.MIDDLE == location:
            if is_header:
                left, mid, right = self._char_hnse, self._char_hnsew, self._char_hnsw
            else:
                left, mid, right = self._char_nse, self._char_nsew, self._char_nsw
                pass
        elif UniTable.BOTTOM == location:
            # NOTE: This will not work as expected if the table is only headers.
            left, mid, right = self._char_ne, self._char_new, self._char_nw
        else:
            raise ValueError("Unknown location '%s'. Should be one of UniTable.TOP, UniTable.MIDDLE, or UniTable.BOTTOM." %(location))
        # compute cell separator
        s = "%s%s%s" % (horiz_char * self._pad, [horiz_char, mid][self.has_vlines()], horiz_char * self._pad)
        # build the line
        l = s.join([horiz_char * n for n in self._width])
        # add border if needed
        if self.has_border:
            l = "%s%s%s%s%s\n" % (left, horiz_char * self._pad , l, horiz_char * self._pad ,right)
        else:
            l += "\n"
        return l

    def _len_cell(self, cell):
        """Return the width of the cell

        Special characters are taken into account to return the width of the
        cell, such like newlines and tabs
        """

        cell_lines = cell.split('\n')
        maxi = 0
        for line in cell_lines:
            length = 0
            parts = line.split('\t')
            for part, i in zip(parts, list(range(1, len(parts) + 1))):
                length = length + len(part)
                if i < len(parts):
                    length = (length//8 + 1) * 8
            maxi = max(maxi, length)
        return maxi

    def _compute_cols_width(self):
        """Return an array with the width of each column

        If a specific width has been specified, exit. If the total of the
        columns width exceed the table desired width, another width will be
        computed to fit, and cells will be wrapped.
        """

        if hasattr(self, "_width"):
            return
        maxi = []
        if self._header:
            maxi = [ self._len_cell(x) for x in self._header ]
        for row in self._rows:
            for cell,i in zip(row, list(range(len(row)))):
                try:
                    maxi[i] = max(maxi[i], self._len_cell(cell))
                except (TypeError, IndexError):
                    maxi.append(self._len_cell(cell))

        ncols = len(maxi)
        content_width = sum(maxi)
        deco_width = 3*(ncols-1) + [0,4][self.has_border]
        if self._max_width and (content_width + deco_width) > self._max_width:
            """ content too wide to fit the expected max_width
            let's recompute maximum cell width for each cell
            """
            if self._max_width < (ncols + deco_width):
                raise ValueError('max_width too low to render data')
            available_width = self._max_width - deco_width
            newmaxi = [0] * ncols
            i = 0
            while available_width > 0:
                if newmaxi[i] < maxi[i]:
                    newmaxi[i] += 1
                    available_width -= 1
                i = (i + 1) % ncols
            maxi = newmaxi
        self._width = maxi

    def _check_align(self):
        """Check if alignment has been specified, set default one if not
        """

        if not hasattr(self, "_header_align"):
            self._header_align = ["c"] * self._row_size
        if not hasattr(self, "_align"):
            self._align = ["l"] * self._row_size
        if not hasattr(self, "_valign"):
            self._valign = ["t"] * self._row_size

    def _draw_line(self, line, isheader=False):
        """Draw a line

        Loop over a single cell length, over all the cells
        """

        line = self._splitit(line, isheader)
        space = " "
        out = ""
        topmost,leftmost = True, True
        for i in range(len(line[0])):
            if self.has_border:
                out += "%s%s" %(self._char_ns, " " * self._pad)
            length = 0
            for cell, width, align in zip(line, self._width, self._align):
                length += 1
                cell_line = cell[i]
                fill = width - len(cell_line)
                if isheader:
                    align = self._header_align[length - 1]
                if align == "r":
                    out += fill * space + cell_line
                elif align == "c":
                    out += (int(fill/2) * space + cell_line \
                            + int(fill/2 + fill%2) * space)
                else:
                    out += cell_line + fill * space
                if length < len(line):
                    out += "%s%s%s" %(" " * self._pad, [space, self._char_ns][self.has_vlines()], " " * self._pad)
            out += "%s\n" % ['', " " * self._pad + self._char_ns][self.has_border]
        return out

    def _splitit(self, line, isheader):
        """Split each element of line to fit the column width

        Each element is turned into a list, result of the wrapping of the
        string to the desired width
        """

        line_wrapped = []
        for cell, width in zip(line, self._width):
            array = []
            for c in cell.split('\n'):
                if c.strip() == "":
                    array.append("")
                else:
                    array.extend(textwrapper(c, width))
            line_wrapped.append(array)
        max_cell_lines = reduce(max, list(map(len, line_wrapped)))
        for cell, valign in zip(line_wrapped, self._valign):
            if isheader:
                valign = "t"
            if valign == "m":
                missing = max_cell_lines - len(cell)
                cell[:0] = [""] * int(missing / 2)
                cell.extend([""] * int(missing / 2 + missing % 2))
            elif valign == "b":
                cell[:0] = [""] * (max_cell_lines - len(cell))
            else:
                cell.extend([""] * (max_cell_lines - len(cell)))
        return line_wrapped



def test_styles(table):
    row = []
    for style in styles:
        table.set_style(style)
        row.append(t2.draw())
        pass
    return row

if __name__ == '__main__':
    table = UniTable()
    table.set_cols_align(["l", "r", "c"])
    table.set_cols_valign(["t", "m", "b"])
    table.add_rows([["Name", "Age", "Nickname"],
                    ["Mr\nXavier\nHuon", 32, "Xav'"],
                    ["Mr\nBaptiste\nClement", 1, "Baby"],
                    ["Mme\nLouise\nBourgeau", 28, "Lou\n \nLoue"]])
    print(table.draw() + "\n")

    table = UniTable()
    table.set_deco(UniTable.HEADER)
    table.set_cols_dtype(['t',  # text
                          'f',  # float (decimal)
                          'e',  # float (exponent)
                          'i',  # integer
                          'a']) # automatic
    table.set_cols_align(["l", "r", "r", "r", "l"])
    table.add_rows([["text",    "float", "exp", "int", "auto"],
                    ["abcd",    "67",    654,   89,    128.001],
                    ["efghijk", 67.5434, .654,  89.6,  12800000000000000000000.00023],
                    ["lmn",     5e-78,   5e-78, 89.4,  .000000000000128],
                    ["opqrstu", .023,    5e+78, 92.,   12800000000000000000000]])
    print(table.draw())

    # Create a table of tables that shows different table styles
    styles = sorted(UniTable.STYLES.keys())
    t1 = UniTable([["STYLES"] + styles])
    t1.set_max_width(0)
    t1.set_cols_align("l" + "c" * len(styles))
    t1.set_cols_valign("m" + "t" * len(styles))
    t1.set_style("light")
    style_rows =[["Header 1","Header 2"],["Cell 1","Cell 2"],["Cell 3","Cell 4"],]
    t2 = UniTable(style_rows)
    for style in styles:
        print("Style \"%s\"" %(style))
        t2.set_style(style)
        print(t2.draw())
        pass
    exit()
    t1.add_row(["DEFAULT"] + test_styles(t2))
    t2.set_padding(0)
    t1.add_row(["set_padding(0)"] + test_styles(t2))
    t2.set_padding(2)
    t1.add_row(["set_padding(2)"] + test_styles(t2))
    t2.set_deco(UniTable.HEADER)
    t2.set_padding(1)
    t1.add_row(["set_deco(HEADER)"] + test_styles(t2))
    print(t1.draw() + "\n")



import setuptools
from md2gemini import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="md2gemini",  # Replace with your own username
    version=__version__,
    author="makeworld",
    author_email="makeworld@protonmail.com",
    description="Convert Markdown to the Gemini text format",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/makeworld-the-better-one/md2gemini",
    packages=["md2gemini"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: General",
        "Topic :: Text Processing :: Markup",
    ],
    python_requires=">=3.7",
    install_requires=[
        "mistune>=2.0.0,<3",
        "cjkwrap",
        "wcwidth",
    ],
    entry_points={"console_scripts": ["md2gemini = md2gemini:main"]},
)



from .util import normalize
from md2gemini import md2gemini


def f(md):
    return normalize(md2gemini(md))


def test_no_extra_newline():
    md = "Non code block\n\n    code block here\n\nMore non code"
    gem = "Non code block\n\n```\ncode block here\n```\n\nMore non code"
    assert f(md) == gem


def test_with_extra_newline():
    md = "Non code block\n\n    code block here\n\n\nMore non code"
    gem = "Non code block\n\n```\ncode block here\n\n```\n\nMore non code"
    assert f(md) == gem


def test_ends_with_no_newlines():
    md = "Non code block\n\n    code block here"
    gem = "Non code block\n\n```\ncode block here\n```"
    assert f(md) == gem


def test_extra_newlines():
    # https://github.com/makeworld-the-better-one/md2gemini/issues/40
    md = "```\ntest\n\n\ntest2\n```\nokay"
    gem = "```\ntest\n\n\ntest2\n```\n\nokay"
    assert f(md) == gem


def test_nested_fences():
    md = """
```
    spaces should be preserved
    ```
    even in here
    there will be two lines
    ```
    and still preserved, as that fence should not end the block
```
And now the block has ended.
This line should be eaten. (note -1 below)
"""
    assert len(md.splitlines()) - 1 == len(f(md).splitlines())


def test_nested_fences_in_lists():
    md = """
1.  This item will have a fence.
    ```
    This is the fence.
    And this line shall be preserved.
        ```begin nested fence
        nested fence
        content
        ```
    Continue preserving newlines.
    ```
2.  This item has no fence
"""
    assert len(md.splitlines()) == len(f(md).splitlines())


def test_nested_fences_in_lists_uneven():
    md = """
1.  This item will have a fence.
    `````
    This is the fence.
    And this line shall be preserved.
        ```begin nested fence
        nested fence
        content
        ````
    Continue preserving newlines.
    ````````````````
2.  This item has no fence
"""
    assert len(md.splitlines()) == len(f(md).splitlines())


def test_nested_fences_in_lists_uneven():
    md = """
1.  This item will have a fence.
    ~~~~~
    This is the fence.
    And this line shall be preserved.
        ```begin nested fence
        nested fence
        content
        ```
    Continue preserving newlines.
    ~~~~~~~~~~~~~~~~
2.  This item has no fence
"""
    assert len(md.splitlines()) == len(f(md).splitlines())
